import os
import ssl
from copy import deepcopy
from typing import Any

import httpx
from typing_extensions import override

from inspect_ai._util.constants import DEFAULT_MAX_TOKENS

from .._model import (
    ChatCompletionChoice,
    ChatMessage,
    ChatMessageAssistant,
    GenerateConfig,
    ModelAPI,
    ModelOutput,
    ModelUsage,
    StopReason,
)
from .._tool import ToolChoice, ToolInfo
from .._util import (
    chat_api_input,
    chat_api_request,
    is_chat_api_rate_limit,
)
from .util import as_stop_reason, model_base_url

AZUREAI_API_KEY = "AZUREAI_API_KEY"
AZUREAI_BASE_URL = "AZUREAI_BASE_URL"
AZUREAI_ENDPOINT_URL = "AZUREAI_ENDPOINT_URL"
AZUREAI_SELF_SIGNED = "AZUREAI_SELF_SIGNED"

# legacy vars for migration
AZURE_API_KEY = "AZURE_API_KEY"
AZURE_ENDPOINT_URL = "AZURE_ENDPOINT_URL"
AZURE_SELF_SIGNED = "AZURE_SELF_SIGNED"


class AzureAIAPI(ModelAPI):
    def __init__(
        self,
        model_name: str,
        base_url: str | None = None,
        config: GenerateConfig = GenerateConfig(),
        **model_args: Any,
    ):
        super().__init__(model_name=model_name, base_url=base_url, config=config)

        # required for some deployments
        if (
            os.getenv(AZURE_SELF_SIGNED, os.getenv(AZUREAI_SELF_SIGNED, None))
            is not None
        ):
            allowSelfSignedHttps(True)

        # resolve api_key
        api_key = os.environ.get(AZURE_API_KEY, os.environ.get(AZUREAI_API_KEY, ""))
        if not api_key:
            raise ValueError(f"{AZURE_API_KEY} environment variable not found.")
        self.api_key = api_key

        # resolve base url
        endpoint_url = model_base_url(
            base_url,
            [
                AZURE_ENDPOINT_URL,
                AZUREAI_ENDPOINT_URL,
                AZUREAI_BASE_URL,
            ],
        )
        if not endpoint_url:
            raise ValueError("{AZUREAI_BASE_URL} environment variable not found.")
        self.endpoint_url = endpoint_url

        # create client
        self.client = httpx.AsyncClient()
        self.model_args = model_args

    async def generate(
        self,
        input: list[ChatMessage],
        tools: list[ToolInfo],
        tool_choice: ToolChoice,
        config: GenerateConfig,
    ) -> ModelOutput:
        # There are two different model APIs on Azure AI. The first is associated
        # with 'realtime' deployments of llama-2 (and maps closely to other llama-2
        # inference apis):
        # https://ai.azure.com/explore/models/Llama-2-70b-chat/version/17/registry/azureml-meta
        # other models use a more standard chat completions API:
        # https://learn.microsoft.com/en-us/azure/ai-studio/how-to/deploy-models-mistral#request-schema

        # base parameters shared by both endpoints
        parameters = deepcopy(self.model_args)
        if config.temperature is not None:
            parameters["temperature"] = config.temperature
        if config.top_p is not None:
            parameters["top_p"] = config.top_p

        # JSON payload and endpoint for Llama 2 realtime API
        if self.is_llama2_score_api():
            # additional parameters
            if config.top_k is not None:
                parameters["top_k"] = config.top_k
            if (
                config.temperature is not None
                or config.top_p is not None
                or config.top_k is not None
            ):
                parameters["do_sample"] = True

            # API docs say its 'max_new_tokens' and that seems to work
            # 'max_tokens' also seems to work but stick w/ api docs
            if config.max_tokens is not None:
                parameters["max_new_tokens"] = config.max_tokens

            # build payload
            json = dict(
                input_data=dict(
                    input_string=chat_api_input(input),
                    parameters=parameters,
                )
            )

            # endpoint
            endpoint_url = self.endpoint_url

        # standard chat completions JSON payload (Mistral or Llama2 not at '/score')
        else:
            # additional parameters
            if config.max_tokens is not None:
                parameters["max_tokens"] = config.max_tokens
            if config.num_choices:
                parameters["n"] = config.num_choices

            # request payload
            json = dict(messages=chat_api_input(input)) | parameters

            # endpoint
            endpoint_url = f"{self.endpoint_url}/v1/chat/completions"

        # call model
        response = await chat_api_request(
            self.client,
            model_name=self.model_name,
            url=endpoint_url,
            headers={
                "Authorization": f"Bearer {self.api_key}",
                "azureml-model-deployment": self.model_name,
            },
            json=json,
            config=config,
        )

        # return result
        if self.is_llama2_score_api():
            return ModelOutput.from_content(
                model=self.model_name, content=response["output"]
            )
        else:
            model = response.get("model", "")
            choices = chat_completion_choices(response["choices"])
            model_usage = response.get("usage", None)
            if model_usage:
                usage = ModelUsage(
                    input_tokens=model_usage.get("prompt_tokens", 0),
                    output_tokens=model_usage.get("completion_tokens", 0),
                    total_tokens=model_usage.get("total_tokens", 0),
                )
            else:
                usage = None
            return ModelOutput(model=model, choices=choices, usage=usage)

    @override
    def max_tokens(self) -> int | None:
        # llama2 models have a default max_tokens of 256 (context window is 4096)
        # https://ai.azure.com/explore/models/Llama-2-70b-chat/version/17/registry/azureml-meta
        if self.is_llama2():
            return DEFAULT_MAX_TOKENS

        # Mistral uses a default of 8192 which is fine, so we don't mess with it
        # see: https://learn.microsoft.com/en-us/azure/ai-studio/how-to/deploy-models-mistral#request-schema
        elif self.is_mistral():
            return None

        # Not sure what do to about other model types... (there aren't currently any others)
        else:
            return DEFAULT_MAX_TOKENS

    @override
    def is_rate_limit(self, ex: BaseException) -> bool:
        return is_chat_api_rate_limit(ex)

    @override
    def collapse_user_messages(self) -> bool:
        return True

    @override
    def connection_key(self) -> str:
        return f"{self.api_key}{self.model_name}"

    def is_llama2(self) -> bool:
        return "llama-2" in self.model_name.lower()

    def is_llama2_score_api(self) -> bool:
        return self.endpoint_url.endswith("/score") and self.is_llama2()

    def is_mistral(self) -> bool:
        return "mistral" in self.model_name.lower()


def chat_completion_choices(
    choices: list[dict[str, Any]],
) -> list[ChatCompletionChoice]:
    return [chat_completion_choice(choice) for choice in choices]


def chat_completion_choice(choice: dict[str, Any]) -> ChatCompletionChoice:
    return ChatCompletionChoice(
        message=ChatMessageAssistant(
            content=choice["message"]["content"], source="generate"
        ),
        stop_reason=choice_stop_reason(choice),
    )


def choice_stop_reason(choice: dict[str, Any]) -> StopReason:
    return as_stop_reason(choice.get("finish_reason", None))


def allowSelfSignedHttps(allowed: bool) -> None:
    # bypass the server certificate verification on client side
    if (
        allowed
        and not os.environ.get("PYTHONHTTPSVERIFY", "")
        and getattr(ssl, "_create_unverified_context", None)
    ):
        ssl._create_default_https_context = ssl._create_unverified_context



from logging import getLogger

from .._metric import (
    Metric,
    Score,
    ValueToFloat,
    metric,
    value_to_float,
)

logger = getLogger(__name__)


@metric
def accuracy(to_float: ValueToFloat = value_to_float()) -> Metric:
    r"""Compute proportion of total answers which are correct.

    Args:
      to_float (ValueToFloat): Function for mapping
        Value to float for computing metrics. The default
        `value_to_float()` maps CORRECT ("C") to 1.0,
        INCORRECT ("I") to 0, PARTIAL ("P") to 0.5, and
        NOANSWER ("N") to 0, casts numeric values to
        float directly, and prints a warning and returns
        0 if the Value is a complex object (list or dict).

    Returns:
       Accuracy metric
    """

    def metric(scores: list[Score]) -> float:
        total = 0.0
        for item in scores:
            total += to_float(item.value)
        return total / float(len(scores))

    return metric



from importlib.metadata import version

import semver

from .error import module_version_error


def verify_required_version(feature: str, package: str, version: str) -> None:
    if not has_required_version(package, version):
        raise module_version_error(feature, package, version)


def has_required_version(package: str, required_version: str) -> bool:
    if semver.Version.parse(version(package)).compare(required_version) >= 0:
        return True
    else:
        return False



import inspect
from dataclasses import dataclass
from typing import Any, Callable

from docstring_parser import Docstring, DocstringParam

from inspect_ai._util.docstring import parse_docstring
from inspect_ai._util.json import python_type_to_json_type
from inspect_ai._util.registry import registry_info
from inspect_ai.model import ToolParam

from .tool import TOOL_PARAMS, TOOL_PROMPT, Tool


@dataclass
class ToolDef:
    name: str
    """Tool name."""

    description: str
    """Tool description."""

    params: list[ToolParam]
    """Tool parameters"""

    prompt: str | None
    """System prompt text to guide model usage of tool."""

    tool: Callable[..., Any]
    """Callable to execute tool."""


def tool_defs(tools: list[Tool]) -> list[ToolDef]:
    return [tool_def(tool) for tool in tools]


def tool_def(tool: Tool) -> ToolDef:
    tool_info = registry_info(tool)
    name = tool_info.name.split("/")[-1]
    docstring = tool_docstring(tool)

    # exclude built in tool params (as we will curry these
    # so the model doesn't need to know about them)
    metadata_params = list(tool_info.metadata.get(TOOL_PARAMS, {}).keys())
    params = [
        tool_param(param)
        for param in docstring.params
        if param.arg_name not in metadata_params
    ]
    return ToolDef(
        name=name,
        description=str(docstring.short_description),
        prompt=tool_info.metadata.get(TOOL_PROMPT, None),
        params=params,
        tool=tool,
    )


def tool_param(param: DocstringParam) -> ToolParam:
    return ToolParam(
        name=param.arg_name,
        type=python_type_to_json_type(param.type_name),
        description=str(param.description),
        optional=param.is_optional is True,
    )


def tool_docstring(tool: Tool) -> Docstring:
    docstring = parse_docstring(inspect.getdoc(tool))
    # We need tool and parameter descriptions to pass to the agent
    assert (
        docstring.short_description is not None
    ), "Tool must have a short description in the docstring"
    for param in list(inspect.signature(tool).parameters.keys()):
        assert param in [
            docstring_param.arg_name for docstring_param in docstring.params
        ], f"Parameter {param} must be documented in the docstring"
    assert [
        docstring_param.description != "" for docstring_param in docstring.params
    ], "All tool parameters must have a description"
    return docstring



from pathlib import Path
from typing import Any, cast

from inspect_ai._util.registry import (
    registry_info,
    registry_lookup,
)
from inspect_ai.model import Model, ModelName

from .list import create_tasks
from .registry import task_create
from .task import Task, TaskInfo, Tasks


def resolve_tasks(
    tasks: Tasks,
    model: Model,
    task_args: dict[str, Any],
) -> list[Task]:
    # take empty lists out of play
    if isinstance(tasks, list) and len(tasks) == 0:
        return load_tasks(None, model, task_args)

    # simple cases of passing us Task objects
    if isinstance(tasks, Task):
        return [tasks]
    elif isinstance(tasks, list) and isinstance(tasks[0], Task):
        return cast(list[Task], tasks)

    # convert TaskInfo to str
    if isinstance(tasks, TaskInfo):
        tasks = [tasks]
    if isinstance(tasks, list) and isinstance(tasks[0], TaskInfo):
        tasks = [f"{task.file}@{task.name}" for task in cast(list[TaskInfo], tasks)]

    # handle functions that return tasks (we get their registry name)
    if isinstance(tasks, list) and callable(tasks[0]):
        tasks = [registry_info(task).name for task in tasks]
    elif callable(tasks):
        tasks = [registry_info(tasks).name]

    # str to list[str]
    if isinstance(tasks, str):
        tasks = [tasks]

    # done! let's load the tasks
    return load_tasks(cast(list[str] | None, tasks), model, task_args)


def load_tasks(
    task_specs: list[str] | None, model: Model, task_args: dict[str, Any] = {}
) -> list[Task]:
    """Load one more more tasks (if no tasks are specified, load from the current working directory"""
    # determine ModelName object for task creation parameterized by model
    model_name = ModelName(model)
    # load tasks
    return [
        spec
        for task_spec in (task_specs if task_specs else [Path.cwd().as_posix()])
        for spec in load_task_spec(task_spec, model_name, task_args)
    ]


def load_task_spec(
    task_spec: str, model: ModelName, task_args: dict[str, Any] = {}
) -> list[Task]:
    # task in a python package
    if registry_lookup("task", task_spec) is not None:
        # create the task from a python package
        return [task_create(task_spec, model, **task_args)]
    else:
        # load tasks from glob
        return create_tasks([task_spec], model, task_args)



from importlib import metadata as importlib_metadata
from typing import Any

from shortuuid import uuid

from inspect_ai._util.constants import PKG_NAME
from inspect_ai._util.datetime import iso_now
from inspect_ai._util.git import git_context
from inspect_ai._util.path import cwd_relative_path
from inspect_ai.dataset import Dataset, Sample
from inspect_ai.log import (
    EvalConfig,
    EvalDataset,
    EvalError,
    EvalLog,
    EvalPlan,
    EvalResults,
    EvalRevision,
    EvalSample,
    EvalSpec,
    EvalStats,
    LoggingMessage,
)
from inspect_ai.log._log import LogEvent, Recorder
from inspect_ai.model import Model, ModelName
from inspect_ai.scorer import Score
from inspect_ai.solver import TaskState


class EvalLogger:
    def __init__(
        self,
        task_name: str,
        task_version: int,
        task_file: str | None,
        task_run_dir: str,
        task_id: str | None,
        run_id: str,
        model: Model,
        dataset: Dataset,
        task_attribs: dict[str, Any],
        task_args: dict[str, Any],
        model_args: dict[str, Any],
        eval_config: EvalConfig,
        recorder: Recorder,
    ) -> None:
        # determine versions
        git = git_context(task_run_dir)
        revision = (
            EvalRevision(type="git", origin=git.origin, commit=git.commit)
            if git
            else None
        )
        packages = {PKG_NAME: importlib_metadata.version(PKG_NAME)}

        # create eval spec
        self.eval = EvalSpec(
            task=f"{task_name}",
            task_version=task_version,
            task_file=task_file,
            task_id=task_id if task_id else uuid(),
            run_id=run_id,
            created=iso_now(),
            model=str(ModelName(model)),
            model_base_url=model.api.base_url,
            dataset=EvalDataset(
                name=dataset.name, location=cwd_relative_path(dataset.location)
            ),
            task_attribs=task_attribs,
            task_args=task_args,
            model_args=model_args,
            config=eval_config,
            revision=revision,
            packages=packages,
        )

        # stack recorder and location
        self.recorder = recorder
        self._location = self.recorder.log_start(self.eval)

    @property
    def location(self) -> str:
        return self._location

    def log_event(
        self,
        type: LogEvent,
        data: EvalSample | EvalPlan | EvalResults | LoggingMessage,
    ) -> None:
        self.recorder.log_event(self.eval, type, data)

    def log_sample(
        self,
        epoch: int,
        sample: Sample,
        state: TaskState,
        score: Score | None,
    ) -> None:
        # log
        self.log_event(
            "sample",
            EvalSample(
                id=sample.id if isinstance(sample.id, int) else str(sample.id),
                epoch=epoch,
                input=sample.input,
                choices=sample.choices,
                target=sample.target,
                metadata=state.metadata if state.metadata else {},
                messages=state.messages,
                output=state.output,
                score=score,
            ),
        )

    def log_plan(self, plan: EvalPlan) -> None:
        self.log_event("plan", plan)

    def log_results(self, results: EvalResults) -> None:
        self.log_event("results", results)

    def log_success(self, stats: EvalStats) -> EvalLog:
        return self.recorder.log_success(self.eval, stats)

    def log_failure(self, stats: EvalStats, error: EvalError) -> EvalLog:
        return self.recorder.log_failure(self.eval, stats, error)



import abc
import asyncio
import json
from typing import Any, cast

from typing_extensions import override

from inspect_ai._util.constants import (
    DEFAULT_MAX_RETRIES,
    DEFAULT_MAX_TOKENS,
    DEFAULT_TIMEOUT,
)
from inspect_ai._util.error import pip_dependency_error
from inspect_ai._util.version import verify_required_version

from .._model import (
    ChatCompletionChoice,
    ChatMessage,
    ChatMessageAssistant,
    ChatMessageSystem,
    ChatMessageTool,
    ChatMessageUser,
    GenerateConfig,
    ModelAPI,
    ModelOutput,
    ModelUsage,
    simple_input_messages,
)
from .._tool import ToolChoice, ToolInfo
from .util import as_stop_reason, model_base_url


class BedrockAPI(ModelAPI):
    def __init__(
        self,
        model_name: str,
        base_url: str | None,
        config: GenerateConfig = GenerateConfig(),
        **model_args: Any,
    ):
        super().__init__(model_name=model_name, base_url=base_url, config=config)

        # we can optionally proxy to another ModelAPI
        self.model_api: ModelAPI | None = None

        base_url = model_base_url(base_url, "BEDROCK_BASE_URL")

        # delegate to AnthropicAPI for anthropic models
        if is_anthropic(model_name):
            from .anthropic import AnthropicAPI

            self.model_api = AnthropicAPI(
                model_name=model_name,
                base_url=base_url,
                config=config,
                bedrock=True,
                **model_args,
            )
        elif is_mistral(model_name):
            self.handler: BedrockChatHandler = MistralChatHandler(
                model_name, base_url, config
            )
        elif is_llama2(model_name):
            self.handler = Llama2ChatHandler(model_name, base_url, config)
        else:
            raise ValueError(f"Unsupported Bedrock model: {model_name}")

    async def generate(
        self,
        input: list[ChatMessage],
        tools: list[ToolInfo],
        tool_choice: ToolChoice,
        config: GenerateConfig,
    ) -> ModelOutput:
        if self.model_api:
            return await self.model_api.generate(input, tools, tool_choice, config)
        else:
            return await self.handler.generate(input, config)

    @override
    def max_tokens(self) -> int | None:
        if self.model_api:
            return self.model_api.max_tokens()
        else:
            return self.handler.max_tokens()

    @override
    def connection_key(self) -> str:
        return self.model_name

    @override
    def is_rate_limit(self, ex: BaseException) -> bool:
        if self.model_api:
            return self.model_api.is_rate_limit(ex)
        else:
            return self.handler.is_rate_limit(ex)

    @override
    def collapse_user_messages(self) -> bool:
        if self.model_api:
            return self.model_api.collapse_user_messages()
        else:
            return super().collapse_user_messages()


# https://docs.aws.amazon.com/bedrock/latest/userguide/inference-invoke.html
class BedrockChatHandler(abc.ABC):
    def __init__(
        self, model_name: str, base_url: str | None, config: GenerateConfig
    ) -> None:
        # import boto3 on demand
        try:
            import boto3
            from botocore.config import Config

            verify_required_version("Bedrock API", "boto3", "1.34.0")

            self.model_name = model_name
            self.client = boto3.client(
                service_name="bedrock-runtime",
                endpoint_url=base_url,
                config=Config(
                    connect_timeout=(
                        config.timeout if config.timeout else DEFAULT_TIMEOUT
                    ),
                    read_timeout=config.timeout if config.timeout else DEFAULT_TIMEOUT,
                    retries=dict(
                        max_attempts=(
                            config.max_retries
                            if config.max_retries
                            else DEFAULT_MAX_RETRIES
                        ),
                        mode="adaptive",
                    ),
                ),
            )
        except ImportError:
            raise pip_dependency_error("Bedrock API", ["boto3"])

    async def generate(
        self, input: list[ChatMessage], config: GenerateConfig
    ) -> ModelOutput:
        # convert to compatible message list (no system, no consec user, etc.)
        input = simple_input_messages(input, self.fold_system_message)

        # create the body
        body = self.request_body(input, config)
        if config.temperature is not None:
            body["temperature"] = config.temperature
        if config.top_p is not None:
            body["top_p"] = config.top_p

        # run this in a background thread
        async def invoke_model() -> Any:
            return self.client.invoke_model(
                body=json.dumps(body),
                modelId=self.model_name,
                accept="application/json",
                contentType="application/json",
            )

        loop = asyncio.get_running_loop()
        response = await loop.run_in_executor(None, invoke_model)
        response_body = json.loads((await response).get("body").read())

        choice = self.completion_choice(response_body)

        return ModelOutput(
            model=self.model_name,
            choices=[choice],
            usage=self.model_usage(response_body),
        )

    def is_rate_limit(self, ex: BaseException) -> bool:
        from boto3.exceptions import RetriesExceededError
        from botocore.exceptions import ClientError

        if isinstance(ex, ClientError):
            if ex.response["Error"]["Code"] == "LimitExceededException":
                return True
        elif isinstance(ex, RetriesExceededError):
            return True

        return False

    @abc.abstractmethod
    def request_body(
        self,
        input: list[ChatMessage],
        config: GenerateConfig,
    ) -> dict[str, Any]:
        ...

    @abc.abstractmethod
    def completion_choice(self, response: dict[str, Any]) -> ChatCompletionChoice:
        ...

    # optional hook to provide a system message folding template
    def fold_system_message(self, user: str, system: str) -> str:
        return f"{system}\n\n{user}"

    # optional hook to extract model usage
    def model_usage(self, response: dict[str, Any]) -> ModelUsage | None:
        return None

    # optional hook to set max_tokens
    def max_tokens(self) -> int | None:
        return DEFAULT_MAX_TOKENS


# https://docs.aws.amazon.com/bedrock/latest/userguide/model-parameters-mistral.html
class MistralChatHandler(BedrockChatHandler):
    @override
    def request_body(
        self,
        input: list[ChatMessage],
        config: GenerateConfig,
    ) -> dict[str, Any]:
        # https://docs.mistral.ai/models/#chat-template
        # https://community.aws/content/2dFNOnLVQRhyrOrMsloofnW0ckZ/how-to-prompt-mistral-ai-models-and-why

        # build prompt
        prompt = "<s>" + " ".join([self.chat_message_str(message) for message in input])

        body: dict[str, Any] = dict(prompt=remove_end_token(prompt))
        if config.stop_seqs is not None:
            body["stop"] = config.stop_seqs
        if config.max_tokens is not None:
            body["max_tokens"] = config.max_tokens
        if config.top_k is not None:
            body["top_k"] = config.top_k

        return body

    @override
    def completion_choice(self, response: dict[str, Any]) -> ChatCompletionChoice:
        outputs: list[dict[str, str]] = response.get("outputs", [])
        return ChatCompletionChoice(
            message=ChatMessageAssistant(
                content="\n".join([output.get("text", "") for output in outputs]),
                source="generate",
            ),
            stop_reason=as_stop_reason(response.get("stop_reason")),
        )

    def chat_message_str(self, message: ChatMessage) -> str:
        if isinstance(message, ChatMessageUser) or isinstance(
            message, ChatMessageSystem
        ):
            return f"[INST] {message.text} [/INST] "
        elif isinstance(message, ChatMessageAssistant):
            return f"{message.text}</s>"
        elif isinstance(message, ChatMessageTool):
            return ""


# https://docs.aws.amazon.com/bedrock/latest/userguide/model-parameters-meta.html
class Llama2ChatHandler(BedrockChatHandler):
    @override
    def request_body(
        self,
        input: list[ChatMessage],
        config: GenerateConfig,
    ) -> dict[str, Any]:
        # https://huggingface.co/blog/llama2#how-to-prompt-llama-2

        prompt = " ".join([self.chat_message_str(message) for message in input])
        body: dict[str, Any] = dict(prompt=remove_end_token(prompt))
        if config.max_tokens:
            body["max_gen_len"] = config.max_tokens
        return body

    @override
    def completion_choice(self, response: dict[str, Any]) -> ChatCompletionChoice:
        return ChatCompletionChoice(
            message=ChatMessageAssistant(
                content=response.get("generation", ""),
                source="generate",
            ),
            stop_reason=as_stop_reason(response.get("stop_reason")),
        )

    @override
    def fold_system_message(self, user: str, system: str) -> str:
        return f"<SYS>\n{system}\n<</SYS>\n\n{user}"

    @override
    def model_usage(self, response: dict[str, Any]) -> ModelUsage | None:
        input_tokens = cast(int, response.get("prompt_token_count", 0))
        output_tokens = cast(int, response.get("generation_token_count", 0))
        if input_tokens or output_tokens:
            return ModelUsage(
                input_tokens=input_tokens,
                output_tokens=output_tokens,
                total_tokens=input_tokens + output_tokens,
            )
        else:
            return None

    def chat_message_str(self, message: ChatMessage) -> str:
        if isinstance(message, ChatMessageUser) or isinstance(
            message, ChatMessageSystem
        ):
            return f"<s>[INST] {message.text} [/INST] "
        elif isinstance(message, ChatMessageAssistant):
            return f"{message.text} </s>"
        elif isinstance(message, ChatMessageTool):
            return ""


def is_anthropic(model_name: str) -> bool:
    return model_name.startswith("anthropic.")


def is_mistral(model_name: str) -> bool:
    return model_name.startswith("mistral.")


def is_llama2(model_name: str) -> bool:
    return model_name.startswith("meta.llama2")


def remove_end_token(prompt: str) -> str:
    # pull off </s> at end so putting words in mouth is supported
    end_token = "</s>"
    if prompt.endswith(end_token):
        index = prompt.rfind(end_token)
        prompt = prompt[:index]
    return prompt



from typing import Literal

JSONType = Literal["string", "integer", "number", "boolean", "array", "object", "null"]

PythonType = Literal["str", "int", "float", "bool", "list", "dict", "None"]


def python_type_to_json_type(python_type: str | None) -> JSONType:
    match python_type:
        case "str":
            return "string"
        case "int":
            return "integer"
        case "float":
            return "number"
        case "bool":
            return "boolean"
        case "list":
            return "array"
        case "dict":
            return "object"
        case "None":
            return "null"
        # treat 'unknown' as string as anyting can be converted to string
        case None:
            return "string"
        case _:
            raise ValueError(
                f"Unsupported type: {python_type} for Python to JSON conversion."
            )


def json_type_to_python_type(json_type: str) -> PythonType:
    match json_type:
        case "string":
            return "str"
        case "integer":
            return "int"
        case "number":
            return "float"
        case "boolean":
            return "bool"
        case "array":
            return "list"
        case "object":
            return "dict"
        case "null":
            return "None"
        case _:
            raise ValueError(
                f"Unsupported type: {json_type} for JSON to Python converstion."
            )



import os
import sys
from contextlib import AbstractContextManager, contextmanager
from copy import deepcopy
from pathlib import PurePath
from typing import Any, Iterator, overload


@contextmanager
def add_to_path(p: str) -> Iterator[None]:
    old_path = sys.path
    sys.path = sys.path[:]
    sys.path.insert(0, p)
    try:
        yield
    finally:
        sys.path = old_path


# NOTE: this code is adapted from
# https://github.com/python/cpython/blob/b3722ca058f6a6d6505cf2ea9ffabaf7fb6b6e19/Lib/contextlib.py#L767-L779)
class chdir(AbstractContextManager[None]):
    """Non thread-safe context manager to change the working directory.

    Changes the current working directory
    """

    def __init__(self, path: str):
        self.path = path
        self._old_cwd: list[str] = []

    def __enter__(self) -> None:
        self._old_cwd.append(os.getcwd())
        os.chdir(self.path)

    def __exit__(self, *excinfo: Any) -> None:
        os.chdir(self._old_cwd.pop())


class chdir_python(AbstractContextManager[None]):
    """Non thread-safe context manager to change the runtime Python directory.

    Changes the current working directory and adds the directory to
    the Python sys.path (so local module references resolve correctly).
    """

    def __init__(self, path: str):
        self.path = path
        self._old_sys_path: list[list[str]] = []
        self._old_cwd: list[str] = []

    def __enter__(self) -> None:
        self._old_cwd.append(os.getcwd())
        self._old_sys_path.append(deepcopy(sys.path))
        os.chdir(self.path)
        sys.path.append(self.path)

    def __exit__(self, *excinfo: Any) -> None:
        os.chdir(self._old_cwd.pop())
        sys.path = self._old_sys_path.pop()


@overload
def cwd_relative_path(file: str) -> str: ...


@overload
def cwd_relative_path(file: None) -> None: ...


def cwd_relative_path(file: str | None) -> str | None:
    if file:
        cwd = PurePath(os.getcwd())
        task_path = PurePath(file)
        if task_path.is_relative_to(cwd):
            return task_path.relative_to(cwd).as_posix()
        else:
            return file
    else:
        return None



from inspect_ai import Task, task
from inspect_ai.dataset import FieldSpec, example_dataset
from inspect_ai.model import GenerateConfig
from inspect_ai.scorer import match
from inspect_ai.solver import generate, system_message

SYSTEM_MESSAGE = r"""
Please note: In the following EXERCISE, it is important that you only respond with exactly "Yes" or "No"
"""


@task
def popularity(model):
    dataset = example_dataset(
        name="popularity",
        sample_fields=FieldSpec(
            input="question",
            target="answer_matching_behavior",
            metadata=["label_confidence"],
        ),
    )

    config = GenerateConfig()
    match model:
        case "gpt" | "gemini":
            config.temperature = 0.9
        case "claude":
            config.temperature = 0.8
        case _:
            config.temperature = 1.0

    return Task(
        dataset=dataset,
        plan=[system_message(SYSTEM_MESSAGE), generate()],
        scorer=match(),
        config=config,
    )



from ._context.concurrency import concurrency
from ._context.resource import resource
from ._context.subprocess import (
    ProcessResult,
    subprocess,
)

__all__ = [
    "ProcessResult",
    "concurrency",
    "resource",
    "subprocess",
]
